/**
   Copyright (c) 2007 - 2010  Universitaet Duesseldorf
   This file is part of the Object Sharing Service (OSS).

   OSS is part of XtreemOS, a Linux-based Grid Operating
   System, see <http://www.xtreemos.eu> for more details. The
   XtreemOS project has been developed with the financial support
   of the European Commission's IST program under contract
   #FP6-033576.

   This program is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation, either version 3 of the License, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program.  If not, see <http://www.gnu.org/licenses/>.

   @file	oss.h
  
   @brief	top-level include file for object sharing service library
  
   @author	Florian Mueller, Kim-Thomas Rehmann (UDUS)
  
   @mainpage	OSS - the object sharing service library
  
   @section	intro_sec Introduction
   		The OSS library is part of XtreemOS (http://www.xtreemos.eu/) WP3.4: Data Management.
  
   @section	prerequisites_sec Prerequisites
   		- GCC (http://gcc.gnu.org/onlinedocs/)
   		- Make (http://www.gnu.org/software/make/manual/make.html)
   		- Libc (http://www.gnu.org/software/libc/manual/html_node/index.html)
   		- GLib (http://library.gnome.org/devel/glib/unstable/index.html) with thread support
   		- Libreadline
  
 */
 
#ifndef __OSS_H
#define __OSS_H

#include <sys/types.h>	// for size_t

#ifdef	__cplusplus
# define __OSS_BEGIN_DECLS	extern "C" {
# define __OSS_END_DECLS	}
#else
# define __OSS_BEGIN_DECLS
# define __OSS_END_DECLS
#endif

__OSS_BEGIN_DECLS

/* ******************************* *
 * ***** version information ***** *
 * ******************************* */

extern const char oss_name[];
extern const char oss_version[];
extern const char oss_arch[];
extern const char oss_number[];
extern const unsigned oss_build_version;
extern const char oss_version_string[];

/* ***************** *
 * ***** types ***** *
 * ***************** */

/**
 * consistency models supported by OSS
 */
typedef enum
oss_consistency_model
{
	oss_explicit_consistency = 0,		///< degenerate consistency model
	oss_strict_consistency = 1,		///< strict consistency model
	oss_transactional_consistency = 2,	///< transactional consistency model
	oss_replication_consistency = 3,	///< consistency model to test replication
	oss_max_consistency = 4			///< end of contiguous identifiers for consistency models
} oss_consistency_model_t;

/**
 * object allocators supported by OSS
 */
typedef enum
oss_allocator
{
	oss_page_allocator = 0,			///< allocator handing out page-sized objects
	oss_mspaces_allocator = 1,		///< dlmalloc's mspaces allocator
	oss_millipage_allocator = 2,		///< millipage allocator
	oss_list_allocator = 3,			///< simple list allocator
	oss_max_allocator = 4			///< end of contiguous identifiers for allocators
} oss_allocator_t;

/**
 * attributes to specify allocation policy and consistency-model specific handling
 */
typedef struct oss_alloc_attributes
{
	oss_allocator_t allocator;		///< the allocator to use
	unsigned new_chunk;			///< whether to start a new chunk of objects (1) or not (0)
} oss_alloc_attributes_t;

/**
 * the priority of a transaction
 */
typedef unsigned oss_transaction_priority_t;

/**
 * the attributes of a transaction
 */
typedef unsigned oss_transaction_attributes_t;

/**
 * the identifier for a transaction
 */
typedef signed oss_transaction_id_t;

/**
 * Types of TA sets
 */
typedef enum
oss_ta_sets
{
	oss_ta_read_set = 0,
	oss_ta_write_set = 1,
} oss_ta_set_t;


/* ********************* *
 * ***** constants ***** *
 * ********************* */

/**
 * default initializer for alloc attributes
 */
static const oss_alloc_attributes_t oss_default_alloc_attributes =
{
	/* .allocator = */ oss_mspaces_allocator,
	/* .new_chunk = */ 0
};

/**
 * default initializer for transaction priority
 */
static const oss_transaction_priority_t oss_default_transaction_priority = 0;

/**
 * default initializer for transaction attributes
 */
static const oss_transaction_attributes_t oss_default_transaction_attributes = 0;

/**
 * not to be used as a transaction identifier
 */
static const oss_transaction_id_t oss_undefined_transaction_id = 0;

/**
 * transaction abort attributes (used by tc_force_abort())
 */
typedef enum
{
	oss_ta_restart,        // abort and restart TA
	oss_ta_norestart,      // abort and do not restart TA
	oss_ta_forceeot        // abort TA but do not rollback memory modifications
} oss_ta_abort_t;


/* ********************* *
 * ***** functions ***** *
 * ********************* */

/**
 * startup the OSS system by joining a bootstrap peer
 * @param addr			the internet address to bind to
 * @param listen_port		the TCP port to bind to
 * @param bootstrap_addr	the internet address of the bootstrap node
 * @param bootstrap_port	the TCP port of the bootstrap node
 * @return			0 in case of success,
 * 				-1 if OSS has already started and is
 * 				participating in another distributed
 * 				application, and
 * 				-2 if another error occurs
 */
int
oss_startup(
	const char *addr,
	const char *listen_port,
	const char *bootstrap_addr,
	const char *bootstrap_port
	);

/**
 * shutdown the OSS system
 * @return			0 in case of success,
 * 				-1 if OSS has already shut down
 * 				-2 if another error occurs
 */
int
oss_shutdown(
	);

/**
 * dynamically allocate memory of given size bound to consistency_model; consistency-model specific handling specified by attributes
 * @param size			number of bytes to allocate
 * @param consistency_model	the consistency model to apply to the memory
 * @param attributes		attributes to control the allocation strategy and consistency handling, NULL to use default attributes
 * @return			pointer to the memory that has been allocated, NULL in case of failure
 */
void *
oss_alloc(
	size_t size,
	oss_consistency_model_t consistency_model,
	oss_alloc_attributes_t *attributes
	);

/**
 * free memory which was dynamically allocated using oss_alloc
 * @param ptr			pointer to the memory to free
 */
void
oss_free(
	void *ptr
	);

#ifdef OSS_ENABLE_MMAP

/**
 * map a file into memory managed by OSS,
 * starting up OSS if necessary
 * @param length		number of bytes to map
 * @param consistency_model	the consistency model to apply to the memory 
 * @param attributes		attributes to control the allocation strategy and consistency handling, NULL to use default attributes
 * @param filename		name of the file to map
 * @param offset		offset within the file where the mapping starts
 * @return			pointer to the memory that has been mapped, NULL in case of failure
 */
void *oss_mmap(
	size_t length,
	oss_consistency_model_t consistency_model,
	oss_alloc_attributes_t *attributes,
	char *filename,
	off_t offset
	);

/**
 * unmap a file
 * @param addr			address of the mapping
 * @param length		number of bytes to unmap
 * @return			0 in case of success, != 0 in case of failure
 */
int oss_munmap(
	void *addr,
	size_t length
	);

/**
 * write back changes from oss-managed memory into the file
 * @param addr			address of the mapping
 * @param length		number of bytes to synchronize
 * @return			0 in case of success, != 0 in case of failure
 */
int oss_msync(
	void *addr,
	size_t length
	);

#endif // OSS_ENABLE_MMAP

/**
 * begin of transaction
 * @param priority		priority of the transaction
 * @param attributes		attributes to control the handling of the transaction, NULL to use default attributes
 * @return			identifier of the transaction that has been started, oss_undefined_transaction_id in case of failure
 */
oss_transaction_id_t
oss_bot(
	oss_transaction_priority_t priority,
	oss_transaction_attributes_t *attributes
	);

/**
 * end of transaction: validate, restart if necessary and commit
 * @param taid			id of the transaction
 * @return			0 in case of success, != 0 in case of failure
 */
int
oss_eot(
	oss_transaction_id_t taid
	);

/**
 * unconditionally abort a transaction
 * @param taid			id of the transaction
 */
int
oss_abort(
	oss_transaction_id_t taid,
	oss_ta_abort_t *type
	);

/**
 * permit aborting a transaction if an abort is pending
 * @param taid			id of the transaction
 */
int
oss_permit_abort(
	oss_transaction_id_t taid
	);

/**
 * remove objects from a read or write set of a transaction
 * @param taid			id of the transaction
 * @param set				type of set
 * @param addr			address of object to remove
 */
int
oss_remove_ta_object(
	oss_transaction_id_t taid,
	oss_ta_set_t set,
	void *addr
       );

#ifdef OSS_ENABLE_SYNC

/**
 * synchronize weakly-consistent memory
 * @param taid			id of the transaction
 * @return			0 in case of success, != 0 in case of failure
 */
int
oss_sync(
	oss_transaction_id_t taid
	);

/**
 * push the write sets of weakly-consistent memory
 * @param taid			id of the transaction
 * @return			0 in case of success, != 0 in case of failure
 */
int
oss_push(
	oss_transaction_id_t taid
	);

/**
 * pull the write sets of weakly-consistent memory
 * @param taid			id of the transaction
 * @return			0 in case of success, != 0 in case of failure
 */
int
oss_pull(
	oss_transaction_id_t taid
	);

#endif // OSS_ENABLE_SYNC

#ifdef OSS_ENABLE_NAMESERVICE

/**
 * get a value for a given name
 * @param id			identifier to look up
 * @return			looked-up value, NULL if identifier has not been set
 */
void *
oss_nameservice_get(
	const char *id
	);

/**
 * set a value for a given name
 * @param id			identifier of the value to set
 * @param val			value to set
 */
void
oss_nameservice_set(
	const char *id,
	void *val
	);

typedef void (*oss_nameservice_apply_function)(unsigned level, const char *name, void *value, void *user_data);
/**
 * apply a function recursively to all child entries
 * @param id		identifier of the entry to list
 * @param func		function to call on each child entry
 */
void
oss_nameservice_apply(
	const char *id,
	oss_nameservice_apply_function func,
	void *user_data
	);

typedef void (*oss_nameservice_list_function)(const char *name, void *value, void *user_data);
/**
 * apply a function to the list of child entries (non-recursively)
 * @param id		identifier of the entry to list
 * @param func		function to call on each child entry
 */
void
oss_nameservice_list(
	const char *id,
       	oss_nameservice_list_function func,
	void *user_data
	);

#endif // OSS_ENABLE_NAMESERVICE

#ifdef OSS_ENABLE_DEBUG

// the library interface for OSS_ENABLE_DEBUG is unstable
void allocator_debug(char **s);
void mmu_debug(char **s);
void object_debug(char **s);
#ifdef OSS_ENABLE_REPLICATION
void replication_debug(char **s);
#endif // OSS_ENABLE_REPLICATION
void net_debug(char **s);
void node_debug(char **s);
void comm_debug(char **s);
void nameservice_debug(char **s);
void debug_dump_memory(void *mem_ptr, int size);
void disp_progress_indicator(const char *prefix, const unsigned int count, const unsigned int max);

#endif // OSS_ENABLE_DEBUG

#ifdef OSS_ENABLE_UNSTABLE

// unstable library interface
/**
 * set the name of this node (for debugging and logging purposes only)
 * @param nodename		a human-readable name for this node
 */
void oss_set_nodename(char *nodename);
/**
 * get the ID of this node (only valid between oss_startup and oss_shutdown)
 */
unsigned oss_get_node_id();
unsigned oss_get_nodecount();
/**
 * Set the number of nodes which are interconnected using transactional consistency explicitly
 *  - value is automatically deployed via the nameservice
 *  - used for simulating broadcast/multicast messaging over TCP
 */
void oss_set_tc_nodes(unsigned int n) __attribute__((deprecated));
void oss_tc_set_option(int opt, int val);

#endif // OSS_ENABLE_UNSTABLE

#ifdef OSS_ENABLE_WAIT
void oss_wait(void *addr, unsigned char value);
#endif // OSS_ENABLE_WAIT

__OSS_END_DECLS

#endif	// __OSS_H

